package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.espec.ESpecLogic;
import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.filefetcher.ExperimentSpecificationFileManager;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.parser.ESpecConstants;
import be.iminds.ilabt.jfed.espec.util.ChModPermissionsUtil;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.espec.util.UploadProgressTracker;
import be.iminds.ilabt.jfed.espec.util.UploadProgressTrackerImpl;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.util.ByteArraySshjFile;
import be.iminds.ilabt.jfed.highlevel.util.ESpecMetaFileCreator;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.KeyUtil;
import com.google.common.base.Charsets;
import javafx.util.Pair;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.sftp.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.security.KeyPair;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.stream.Collectors;

/**
 * ExperimentSpecificationServicesDirsAndUploadsState does 3 things:
 *   - Setup SSH connections to nodes, for the dirs, upload, and execute steps
 *   - Do the "dir" step of the ExperimentSpecification
 *   - Do the "upload" step of the ExperimentSpecification
 *
 * Note that these 2 steps can be combined here easily, as there are no inter node dependencies.
 * For the execute step, things are different. See ExperimentSpecificationServicesSingleExecuteState
 *
 * Additionally, this class can be called to close all SSH connections. (typically after all executes)
 *   See TODO
 */
public class ExperimentSpecificationServicesDirsAndUploadsState
        extends SlicedState<ExperimentSpecificationServicesDirsAndUploadsState.ExperimentSpecificationServiceDirsAndUploadsStateSlice> {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentSpecificationServicesDirsAndUploadsState.class);
    @Nonnull
    private final SetupSoftwareExperimentJob setupSoftwareExperimentJob;
    @Nonnull
    private final Experiment experiment;

    @Nonnull
    private final GeniUser geniUser;

    @Nonnull
    private final ExperimentSpecification experimentSpecification;
    @Nonnull
    private final ESpecBundle experimentSpecificationBundle;
    @Nonnull
    private final ESpecLogger eSpecLogger;

    @Nonnull
    private final RemotePathResolver remotePathResolver;

    private final AtomicBoolean anyDirFailed = new AtomicBoolean(false);
    private final AtomicBoolean anyUploadFailed = new AtomicBoolean(false);

    @Nonnull
    private final List<ExperimentSpecificationServiceDirsAndUploadsStateSlice> slices;

    protected ExperimentSpecificationServicesDirsAndUploadsState(@Nonnull SetupSoftwareExperimentJob job,
                                                                 @Nonnull RemotePathResolver remotePathResolver,
                                                                 @Nonnull GeniUser geniUser) {
        super("Execute ExperimentSpecification Dir And Upload Steps");
        this.setupSoftwareExperimentJob = job;
        this.remotePathResolver = remotePathResolver;
        this.geniUser = geniUser;
        this.experiment = job.getExperiment();
        this.eSpecLogger = experiment.getExperimentSpecificationLogger();

        experiment.requireSlice();
        if (experiment.getExperimentSpecification() == null)
            throw new IllegalStateException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification");
        if (experiment.getExperimentSpecificationBundle() == null)
            throw new IllegalStateException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification (bundle)");

        this.experimentSpecification = experiment.getExperimentSpecification();
        this.experimentSpecificationBundle = experiment.getExperimentSpecificationBundle();

        assert experiment.getKeypairs() != null;

        this.slices = new ArrayList<>();


        RequestRspecSource requestRspecSource = experiment.getSlice().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            LOG.warn("ExperimentSpecificationServicesDirsAndUploadsState will do nothing because there is no usable request RSpec");
            return;
        }

        for (FXRspecNode node : model.getNodes()) {
            String nodeId = node.getClientId();

            if (nodeId != null) {
                //Note: Since this needs to setup SSH connections for execute as well, this will also run on nodes with only an execute step.

                Collection<String> ansibleGroups = node.getAnsibleGroups();

                boolean neededOnNode = experimentSpecification.getDirs().stream()
                        .anyMatch(s -> ESpecLogic.mustRunOn(nodeId, ansibleGroups, experimentSpecification, s));
                neededOnNode |= experimentSpecification.getExecutes().stream()
                        .anyMatch(s -> ESpecLogic.mustRunOn(nodeId, ansibleGroups, experimentSpecification, s));
                neededOnNode |= experimentSpecification.getUploads().stream()
                        .anyMatch(s -> ESpecLogic.mustRunOn(nodeId, ansibleGroups, experimentSpecification, s));

                //we assume that there is always an upload needed on the ansible node
                if (ESpecLogic.isAnsibleNode(nodeId, experimentSpecification))
                    neededOnNode = true;

                if (neededOnNode) {
                    slices.add(new ExperimentSpecificationServiceDirsAndUploadsStateSlice(job, node));
                }
            }
        }
    }

    @Nonnull
    @Override
    protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
        //dir and upload cannot be seperated at this level
        //TODO: separate them (if it's worth the trouble)
        eSpecLogger.firePreDir();
        eSpecLogger.firePreUpload();
        ExperimentTaskStatus res;
        try {
            res = super.executeState(job);
        } catch (Throwable t) {
            LOG.error("Failing due to exception in super.executeState", t);
            res = ExperimentTaskStatus.FAILED;
            throw t;
        } finally {
            eSpecLogger.firePostDirAll(!anyDirFailed.get());
            eSpecLogger.firePostUploadAll(!anyUploadFailed.get());
        }
        eSpecLogger.firePreExecute();
        return res;
    }


    @Override
    public Collection<ExperimentSpecificationServiceDirsAndUploadsStateSlice> getSlices() {
        return slices;
    }

    public class ExperimentSpecificationServiceDirsAndUploadsStateSlice extends StateSlice {
        @Nonnull
        protected final FXRspecNode node;
        @Nonnull
        protected final String clientId;
        @Nullable
        private SSHClient ssh;

        @Nullable
        private ExternalFileUtil externalFileUtil;

        public ExperimentSpecificationServiceDirsAndUploadsStateSlice(@Nonnull Job<?> job,
                                                                      @Nonnull FXRspecNode node) {
            super(job);
            this.node = node;
            if (node.getClientId() == null)
                throw new IllegalStateException("Only nodes with a client_id can be used in an ExperimentSpecification");
            this.clientId = node.getClientId();
        }

        @Override
        public String getName() {
            return "ExperimentSpecification steps for node " + clientId;
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            try {
                ssh = setupSoftwareExperimentJob.getSSHClient(node, this);
            } catch (SetupSoftwareExperimentJob.SshException e) {
                LOG.warn("Aborting ExperimentSpecificationServiceDirsAndUploadsStateSlice.statefulRun due to SSH connection problem", e);
                return ExperimentTaskStatus.FAILED;
            }
            assert ssh != null;
            externalFileUtil = new ExternalFileUtil(ssh);

            eSpecLogger.firePreDir(clientId);
            //TODO: only start when needed
            final SetupDirsState setupDirState = new SetupDirsState();
            try {
                setAndRunState(setupDirState);
            } finally {
                if (setupDirState.getStatus() != ExperimentTaskStatus.SUCCESS)
                    anyDirFailed.set(true);
                eSpecLogger.firePostDirNode(clientId, setupDirState.getStatus() == ExperimentTaskStatus.SUCCESS);
            }
            if (setupDirState.getStatus() != ExperimentTaskStatus.SUCCESS) {
                LOG.error("Failing because setupDirState.getStatus()="+setupDirState.getStatus());
                return ExperimentTaskStatus.FAILED;
            }


            eSpecLogger.firePreUpload(clientId);
            final UploadState uploadState = new UploadState();
            try {
                setAndRunState(uploadState);
            } finally {
                if (uploadState.getStatus() != ExperimentTaskStatus.SUCCESS)
                    anyUploadFailed.set(true);
                eSpecLogger.firePostUploadNode(clientId, uploadState.getStatus() == ExperimentTaskStatus.SUCCESS);
            }

            if (uploadState.getStatus() != ExperimentTaskStatus.SUCCESS) {
                LOG.error("Failing because uploadState.getStatus()=" + uploadState.getStatus());
                return ExperimentTaskStatus.FAILED;
            }

            return ExperimentTaskStatus.SUCCESS;
        }

        public class SetupDirsState extends State {
            protected SetupDirsState() {
                super("Checking (and possible creating) dirs on " + node.getClientId());
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                if (ssh == null)
                    throw new IllegalStateException("ssh == null");
                if (externalFileUtil == null)
                    throw new IllegalStateException("externalFileUtil == null");

                List<DirSpec> dirs = new ArrayList<>(experimentSpecification.getDirs());

                for (DirSpec dirSpec : dirs) {
                    if (ESpecLogic.mustRunOn(clientId, node.getAnsibleGroups(), experimentSpecification, dirSpec)) {
                        String remotePath = remotePathResolver.getRemotePath(dirSpec, clientId);
                        long logId = eSpecLogger.getNextLogId();

                        int targetDirMask = ChModPermissionsUtil.toMask(dirSpec.getPermissions());
                        FileAttributes targetDirAtts = new FileAttributes.Builder()
                                .withPermissions(targetDirMask)
                                .withType(FileMode.Type.DIRECTORY).build();

                        if (dirSpec.getDirContent() != null) {
                            assert remotePath != null;
                        }

                        eSpecLogger.firePreDir(logId, dirSpec, remotePath, clientId);

                        try (SFTPClient sftpClient = ssh.newSFTPClient()) {
                            FileAttributes att = sftpClient.statExistence(remotePath);

                            LOG.debug("ExperimentSpecification.dir checked dir {} on {}. exists={}",
                                    remotePath, node.getClientId(), att != null);

                            if (att != null) {
                                //handle existing dir
                                if (att.getType() != FileMode.Type.DIRECTORY) {
                                    throw new IllegalStateException("Error while checking dir "+remotePath+", it already exists but is not a directory");
                                }

                                //don't check permissions, just set them
                                sftpClient.chmod(remotePath, targetDirMask);
                            } else {
                                //create dir(s)
                                SFTPEngine engine = sftpClient.getSFTPEngine();
                                final Deque<String> dirsToMake = new LinkedList<>();
                                for (PathComponents current = engine.getPathHelper().getComponents(remotePath); ;
                                     current = engine.getPathHelper().getComponents(current.getParent())) {
                                    final FileAttributes attrs = sftpClient.statExistence(current.getPath());
                                    if (attrs == null) {
                                        dirsToMake.push(current.getPath());
                                    } else if (attrs.getType() != FileMode.Type.DIRECTORY) {
                                        throw new IllegalStateException("Error while creating dir "+remotePath+" -> "+current.getPath() + " exists but is not a directory");
                                    } else {
                                        break;
                                    }
                                }
                                LOG.debug("ExperimentSpecification.dir must create dirs {}", dirsToMake);
                                while (!dirsToMake.isEmpty()) {
                                    String dirToMake = dirsToMake.pop();
                                    LOG.debug("ExperimentSpecification.dir creating {} on {}",
                                            dirToMake, node.getClientId());
                                    //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work...
                                    engine.makeDir(dirToMake, targetDirAtts);
                                    //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work...
                                    sftpClient.chmod(dirToMake, targetDirMask);
                                }
                            }

                            //TODO: we should set appropriate permissions along the parts of the directory path that already exist. ("execute" perm for uog depending if they get any permissions on final dir)

                            eSpecLogger.firePostDirSuccess(logId, dirSpec, remotePath, clientId);
                        } catch (IOException ex) {
                            LOG.error("Error while creating dir {} on {}", remotePath, node.getClientId(), ex);
                            updateMessage(String.format("Error while creating dir %s on %s: %s", remotePath, node.getClientId(), ex.getMessage()));
                            eSpecLogger.firePostDirFail(logId, dirSpec, remotePath, clientId,
                                    "Error while creating dir "+remotePath+" on "+node.getClientId(), ex);
                            return ExperimentTaskStatus.FAILED;
                        }
                    }
                }

                return ExperimentTaskStatus.SUCCESS;
            }
        }

        public class UploadState extends State {
            protected UploadState() {
                super("Uploading files on " + node.getClientId());
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                if (ssh == null)
                    throw new IllegalStateException("ssh == null");
                if (externalFileUtil == null)
                    throw new IllegalStateException("externalFileUtil == null");

                //We need to wait for all files to be ready
                //noinspection ConstantConditions
                try {
                    job.getExperiment().getExperimentSpecificationFileManager().waitUntilReady();
                    if (job.getExperiment().getExperimentSpecificationFileManager().getErrorOccured())
                        throw new RuntimeException("Error fetching ESpec resources for upload");
                } catch (Exception e) {
                    LOG.error("Error fetching ESpec resources for upload", e);
                    updateMessage("Error fetching ESpec resources for upload: "+e.getMessage());
                    return ExperimentTaskStatus.FAILED;
                }

                //uploads file AND scripts (but don't execute scripts yet)
                List<UploadLikeSpec> posUploads = new ArrayList<>(experimentSpecification.getUploads());
                posUploads.addAll(experimentSpecification.getExecutes());

                List<UploadLikeSpec> ansibleUploads = new ArrayList<>();
                if (experimentSpecification.getAnsible() != null) {
                    ansibleUploads.addAll(experimentSpecification.getAnsible().getAnsibleHostSpec().getUploadLists());
                    ansibleUploads.addAll(experimentSpecification.getAnsible().getAnsibleHostSpec().getExecuteLists());
                    ansibleUploads.addAll(experimentSpecification.getAnsible().getAnsibleGalaxySpecs());
                    ansibleUploads.addAll(experimentSpecification.getAnsible().getAnsiblePlaybookSpecs());
                    for (AnsiblePlaybookSpec ansiblePlaybookSpec : experimentSpecification.getAnsible().getAnsiblePlaybookSpecs()) {
                        if (ansiblePlaybookSpec.hasExtraVarsJsonFileSpec() &&
                                ansiblePlaybookSpec.getExtraVarsJsonFileSpec() != null) {
                            ansibleUploads.add(ansiblePlaybookSpec.getExtraVarsJsonFileSpec());
                        }
                    }
                    posUploads.addAll(ansibleUploads);
                }

                List<UploadLikeSpec> localFiles = posUploads.stream()
                        .filter(f -> f.getSource() == null)
                        .collect(Collectors.toList());

                //exclude files that are already on the node
                List<UploadLikeSpec> uploads = posUploads.stream()
                        .filter(f -> f.getSource() != null)
                        .collect(Collectors.toList());


                LOG.debug("ExperimentSpecification.upload There are "+uploads.size()+" uploads needed.");
                LOG.debug("ExperimentSpecification.upload There are "+localFiles.size()+" files already present on the nodes.");
                LOG.debug("ExperimentSpecification.upload Of these there are "+ansibleUploads.size()+" ansible file uploads or already present files.");
                LOG.debug("ExperimentSpecification.upload Files to be uploaded: "+ uploads.stream()
                        .map(UploadLikeSpec::getDesc)
                        .collect(Collectors.joining(", ")));
                LOG.debug("ExperimentSpecification.upload Files already present: "+ localFiles.stream()
                        .map(UploadLikeSpec::getDesc)
                        .collect(Collectors.joining(", ")));

                //upload the files that need to be uploaded, and register their remote path
                try (SFTPClient sftpClient = ssh.newSFTPClient()) {
                    for (UploadLikeSpec uploadSpec : uploads) {
                        long logId = eSpecLogger.getNextLogId();
                        if (uploadSpec.getSource() == null) {
                            throw new RuntimeException("Internal assumption failed");
                        }

                        if (!ESpecLogic.mustRunOn(clientId, node.getAnsibleGroups(), experimentSpecification, uploadSpec)) {
                            LOG.debug("ExperimentSpecification.upload skips {} on {}  (ansNode={} ansSub={})",
                                    uploadSpec.getDesc(),
                                    clientId,
                                    ESpecLogic.isAnsibleNode(clientId, experimentSpecification),
                                    ESpecLogic.isAnsibleSubPart(uploadSpec, experimentSpecification));
                            continue;
                        }

                        assert externalFileUtil != null;
                        //use dir with "scripts" or "uploads" type as appropriate, fall back to homedir
                        String remotePath = remotePathResolver.getRemotePath(uploadSpec, clientId);
                        assert remotePath != null;

                        UploadProgressTracker uploadProgressTracker = new UploadProgressTrackerImpl();
                        eSpecLogger.firePreUpload(logId, uploadSpec, remotePath, clientId, uploadProgressTracker);

                        String name = externalFileUtil.getBasename(remotePath);

                        try {

                            boolean isDir = job.getExperiment().getExperimentSpecificationFileManager().isDir(uploadSpec.getSource());
                            if (isDir) {
                                if (!remotePath.endsWith("/"))
                                    remotePath += "/";

                                LOG.debug("ExperimentSpecification.upload starts recursive dir upload from {} to {} on {}",
                                        uploadSpec.getSource().getValue(), remotePath, node.getClientId());

                                {
                                    int totalByteCount = 0;
                                    Iterator<ExperimentSpecificationFileManager.DirFileBytes> filesToUpload =
                                            job.getExperiment().getExperimentSpecificationFileManager().getDirBytes(uploadSpec.getSource());
                                    while (filesToUpload.hasNext()) {
                                        ExperimentSpecificationFileManager.DirFileBytes dirFileBytes = filesToUpload.next();
                                        if (!dirFileBytes.isDir()) {
                                            totalByteCount += dirFileBytes.getContent().length;
                                        }
                                    }
                                    uploadProgressTracker.setTotalBytes(totalByteCount);
                                }

                                int currentUploadedByteCount = 0;
                                Iterator<ExperimentSpecificationFileManager.DirFileBytes> filesToUpload =
                                        job.getExperiment().getExperimentSpecificationFileManager().getDirBytes(uploadSpec.getSource());
                                //note: filesToUpload will start with the dir itself (which is convenient)

                                while (filesToUpload.hasNext()) {
                                    ExperimentSpecificationFileManager.DirFileBytes dirFileBytes = filesToUpload.next();

                                    String curRemotePath = remotePath + dirFileBytes.getFullRelativePath();

                                    if (dirFileBytes.isDir()) {
                                        // We can't use uploadSpec.getPermissions()   here, as the parser takes a default,
                                        // but doesn't know at that point this is a dir
                                        // So we use fixed permissions for now. Not ideal there is no clean solution.
                                        // Users that wan't some fancy permissions for the dir, should chmod these themself
                                        makeDir(sftpClient, curRemotePath, "u=rwx");
                                    } else {
                                        //upload a file (directory already exists)
                                        uploadFile(sftpClient,
                                                curRemotePath,
                                                dirFileBytes.getBaseFilename().isEmpty() ?
                                                        name :
                                                        dirFileBytes.getBaseFilename(),
                                                uploadSpec.getPermissions(),
                                                dirFileBytes.getContent());

                                        currentUploadedByteCount += dirFileBytes.getContent().length;
                                        uploadProgressTracker.setCurrentBytes(currentUploadedByteCount);
                                    }

                                    LOG.debug("ExperimentSpecification.upload finished recursive dir upload");
                                }
                            } else {
                                LOG.debug("ExperimentSpecification.upload starts single file upload from {} to {} on {}",
                                        uploadSpec.getSource().getValue(), remotePath, node.getClientId());

                                byte[] content;
                                if (uploadSpec.getSource().getType() == FileSource.SourceType.META) {
                                    content = getMetaContent(uploadSpec.getSource().getValue(), node);
                                }else {
                                    if (uploadSpec.getSource().isGeneratedKeyPair()) {
                                        content = getGeneratedContent(uploadSpec.getSource().getValue());
                                    } else {
                                        content = job.getExperiment().getExperimentSpecificationFileManager().getBytes(uploadSpec.getSource());
                                    }
                                }

                                uploadProgressTracker.setTotalBytes(content.length);

                                uploadFile(sftpClient,
                                        remotePath,
                                        name,
                                        uploadSpec.getPermissions(),
                                        content);

                                //TODO: update intermediate progress
                                uploadProgressTracker.setCurrentBytes(content.length);
                            }

                            eSpecLogger.firePostUploadSuccess(logId, uploadSpec, remotePath, clientId);
                        } catch (IOException ex) {
                            LOG.error("Error while uploading {} to {} on {}", name, remotePath, node.getClientId(), ex);
                            updateMessage(String.format("Error while uploading %s to %s on %s: %s", name, remotePath, node.getClientId(), ex.getMessage()));
                            eSpecLogger.firePostUploadFail(logId, uploadSpec, remotePath, clientId,
                                    "Error while uploading "+name+" to "+remotePath+" on "+node.getClientId(), ex);
                            return ExperimentTaskStatus.FAILED;
                        }
                    }
                } catch (IOException e) {
                    throw new JFedException("Failed to create SSH SFTP client", e);
                }

                return ExperimentTaskStatus.SUCCESS;
            }

            private void makeDir(SFTPClient sftpClient, String remotePath, String dirPerm) throws IOException {
                SFTPEngine engine = sftpClient.getSFTPEngine();

                int targetDirMask = ChModPermissionsUtil.toMask(dirPerm);
                FileAttributes targetDirAtts = new FileAttributes.Builder()
                        .withPermissions(targetDirMask)
                        .withType(FileMode.Type.DIRECTORY).build();

                LOG.debug("ExperimentSpecification.upload creating dir {} on {}",
                        remotePath, node.getClientId());
                //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not always work...
                engine.makeDir(remotePath, targetDirAtts);
                //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not always work...
                sftpClient.chmod(remotePath, targetDirMask);
            }

            private void uploadFile(SFTPClient sftpClient, String remotePath, String name, String perm, byte[] content) throws IOException {
                int targetFileMask = ChModPermissionsUtil.toMask(perm);

                LOG.debug("ExperimentSpecification.upload is uploading {} ({} bytes) to {} on {} with perm={}",
                        name, content.length, remotePath, node.getClientId(), perm);
                //TODO   NOTE: sftpClient.chmod and other permission related commands/parameters in sshj SFTP do not always work...
                sftpClient.getFileTransfer().upload(
                        new ByteArraySshjFile(name, content, perm),
                        remotePath);
                //TODO   NOTE: sftpClient.chmod and other permission related commands/parameters in sshj SFTP do not always work...
                //             u=rx seems to work, but oug=rw does not ?  (but needs testing)
                //             (also not know if when it works, if upload or chmod works or both)
                sftpClient.chmod(remotePath, targetFileMask);
            }


            private byte[] getGeneratedContent(String generatedContentId) {
                if (ESpecConstants.isPrivateKey(generatedContentId)) {
                    if (experiment.getKeypairs() == null || experiment.getKeypairs().isEmpty())
                        throw new IllegalStateException("\"generated\" source \"" + generatedContentId + "\" used, but experiment has no keypairs");
                    KeyPair keyPair = experiment.getKeypairs().values().iterator().next();
                    assert keyPair != null;
                    assert keyPair.getPrivate() != null;
                    char[] privateKeyChars = KeyUtil.privateKeyToPem(keyPair.getPrivate());
                    return new String(privateKeyChars).getBytes(StandardCharsets.UTF_8);
                }
                if (ESpecConstants.isPublicKey(generatedContentId)) {
                    if (experiment.getKeypairs() == null || experiment.getKeypairs().isEmpty())
                        throw new IllegalStateException("\"generated\" source \"" + generatedContentId + "\" used, but experiment has no keypairs");
                    KeyPair keyPair = experiment.getKeypairs().values().iterator().next();
                    assert keyPair != null;
                    assert keyPair.getPublic() != null;
                    String publicKeyString = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(keyPair.getPublic());
                    return publicKeyString.getBytes(StandardCharsets.UTF_8);
                }
                throw new UnsupportedOperationException("Unsupported \"generated\" source \""+generatedContentId+"\"");
            }

            private byte[] getMetaContent(@Nonnull String metaContentId, @Nonnull RspecNode node) {
                ESpecMetaFileCreator metaFileCreator = new ESpecMetaFileCreator(experiment, geniUser);
                if (metaContentId.equals(ESpecConstants.META_MANIFEST)) {
                    return metaFileCreator.getManifestXml();
                }
                if (metaContentId.equals(ESpecConstants.META_EXPERIMENT_INFO_JSON)) {
                    return metaFileCreator.getExperimentInfoJson();
                }
                if (metaContentId.equals(ESpecConstants.META_CLIENT_ID)) {
                    if (node == null)
                        return "INTERNAL ERROR: Unknown node".getBytes(Charsets.UTF_8);
                    if (node.getClientId() == null)
                        return "INTERNAL ERROR: Unknown client_id for node".getBytes(Charsets.UTF_8);
                    return node.getClientId().getBytes(Charsets.UTF_8);
                }
                throw new UnsupportedOperationException("Unsupported \"meta\" value \""+metaContentId+"\"");
            }
        }
    }
}
